package pak;

import java.util.*;
import java.util.zip.*;
import java.io.*;

public class Mappak	// the bsp file header and pak data
{
    int ident;		// bsp ident
    int version;	// bsp version
    Lump [] lump;	// the lumps
    int maprev;		// bsp map revision
    int offset;		// pak offset (from Lump 40)
    int length;		// pak length (from Lump 40)
    
//JDK150    ArrayList<Zipf> zf;		// the zip files in the pak
    ArrayList zf;		// the zip files in the pak
    

    public void savemap(RandomAccessFile bin, RandomAccessFile bout) throws IOException		// write bytes from bin to bout up to pak lump
    {
        for(int i=0;i<offset;i++)			// for length of file minus pak
        {
            bout.writeByte(bin.readUnsignedByte());		// write bytes from buffer to file
        }
    }
    
    public void savepak(RandomAccessFile bin, RandomAccessFile bout) throws IOException
    {
        int numzips = zf.size();		// this number of files to write to pak
        
        for(int i=0;i<numzips;i++)        // write the zip files - localheader+data first
        {
  //    	System.out.print("ZLH "+i);
//JDK150	      Zipf z = zf.get(i);         // the i'th zip
        	Zipf z = (Zipf) zf.get(i);         // the i'th zip
    
            z.relofs = (int)bout.getFilePointer()-offset;   // get rel offset of local header
            					// must Swab the endianness - ugly!
            bout.writeInt(Swab.I(0x04034b50));     // ZLFH signature
            bout.writeShort(Swab.S(10));   // version needed to extract
            bout.writeShort(0);    // flags
            bout.writeShort(0);    // compression method (none)
            bout.writeShort(0);    // last mod time
            bout.writeShort(0);    // last mod date
            bout.writeInt(Swab.I((int)z.CRC));     // CRC32
            bout.writeInt(Swab.I(z.size));          // compressed size
            bout.writeInt(Swab.I(z.size));          // uncompressed size
            bout.writeShort(Swab.S(z.fullname.length()));       // file name length
            bout.writeShort(0);    // exfield length
            writestr(bout, z.fullname);  // the filename
                                          // no exfield

            z.datofs = (int)bout.getFilePointer()-offset;   // get rel offset of zip data

//		System.out.print(":");
            if(z.inpak)                 // in the pak?
            {                  
//			System.out.print("I");
                bin.seek(offset+z.datofs);  // seek to beginning of zip data
                for(int j=0;j<z.size;j++)       // read-write bytes
                    bout.writeByte(bin.readUnsignedByte());
            }
            else
            {       // not in the pak
//			System.out.print("O");
                	bout.write(z.data);           // emit the data buffer
			// file now in pak if it was not, so clean up the internal flags
	  		z.data = null;	// free the buffer
			z.inpak = true;	// is now in pak
            }
//		System.out.println("");
	  }
        
	  
        // end of localheader + data - now write the Central Directory
            
//	  System.out.println("SOCD");
        int cdoffs = (int)bout.getFilePointer()-offset;    // relative offset of beginning of central directory
            
		
        for(int i=0;i<numzips;i++)    // i'th zipfileheader
        {
//JDK150            Zipf z = zf.get(i);     // the i'th zip
            Zipf z = (Zipf) zf.get(i);     // the i'th zip
//		System.out.println("ZFH "+i);
            bout.writeInt(Swab.I(0x02014b50));     // ZFH signature
            bout.writeShort(Swab.S(20));   // version made by
            bout.writeShort(Swab.S(10));   // version needed to extract
            bout.writeShort(0);    // flags
            bout.writeShort(0);    // compression method (none)
            bout.writeShort(0);    // last mod time
            bout.writeShort(0);    // last mod date
            bout.writeInt(Swab.I((int)z.CRC));     // CRC32
            bout.writeInt(Swab.I(z.size));          // compressed size
            bout.writeInt(Swab.I(z.size));          // uncompressed size
            bout.writeShort(Swab.S(z.fullname.length()));       // file name length
            bout.writeShort(0);    // exfield length
            bout.writeShort(0);    // comment length
            bout.writeShort(0);    // disk number start
            bout.writeShort(0);    // internal file attrs
            bout.writeInt(0);              // external file attrs
            bout.writeInt(Swab.I(z.relofs));       // relative offset of local header
            
            writestr(bout,z.fullname);     // filename
                                            // no exfield or comment
        }
        
		// central directory entries all written, now write the end of CD marker
//	System.out.println("EOCD");
        int cdend = (int)bout.getFilePointer() - offset;    // relative offset of end of CD
        
        bout.writeInt(Swab.I(0x06054b50));        // EOCD signature
        bout.writeShort(0);        // number of this disk
        bout.writeShort(0);        // number of disk with SOCD
        bout.writeShort(Swab.S(numzips));  // number of files this disk
        bout.writeShort(Swab.S(numzips));  // number of files total
        bout.writeInt(Swab.I(cdend - cdoffs));     // size of CD
        bout.writeInt(Swab.I(cdoffs));             // offset to start of CD
        bout.writeShort(0);        // comment length
                                          // no comment
            
//	System.out.println("ULL");
        lump[40].len = (int)bout.getFilePointer() - offset;    // compute the new length of the pak lump
//        System.out.println(lump[40].len);
						 // seek to postion of lump len in 40'th lump header
        bout.seek(8 + 640L +4);      // header + 40*lumphsize + 2nd int in lumph
        
        bout.writeInt(Swab.I(lump[40].len));	// write it
        							// all done
//	System.out.println("EOO");
    }
    
    
    public void loadmap(RandomAccessFile b) throws IOException		//Load the map file.
    {
        loadheader(b);		// load the header + lumps
        
        loadpak(b);			// load the pakfile headers
        
    }
    
    
    public void loadheader(RandomAccessFile bb)	throws IOException	// load the map header, parsing lumps
    {
        
        ident = Swab.I(bb.readInt());
        
        version = Swab.I(bb.readInt());
        
        int vbsp = (('P'<<24)+('S'<<16)+('B'<<8)+'V');
        
        if(ident!=vbsp)					// check for correct file format
        {
            Cons.println("Not a map file: "+vbsp+" != "+ident);
            return;
        }
        Cons.println("Ident: "+ident);
        Cons.println("Version: "+version);
        
        lump = new Lump[64];
        
        for(int i=0;i<64;i++)				// read each lump
        {
            lump[i] = new Lump();
            
            lump[i].ofs=Swab.I(bb.readInt());
            lump[i].len=Swab.I(bb.readInt());
            lump[i].vers=Swab.I(bb.readInt());
            lump[i].fourCC=Swab.I(bb.readInt());
            
            if(lump[i].len>0)
            {
                Cons.print("Lump "+i+": ");
                Cons.print(lump[i].ofs+", "+lump[i].len+", "+lump[i].vers+", "+lump[i].fourCC);
                Cons.println(" "+Lump.name(i)+"  "+lump[i].len/Lump.size(i)+(Lump.size(i)==1?" bytes":""));
            }
        }
        
        maprev = Swab.I(bb.readInt());
        Cons.println("MapRev: "+maprev+"\n----------------\n");
        
        //Check Pak lump is last
        
        int last = 0;
        
        for(int i=0;i<64;i++)
        {
            if(lump[i].ofs > lump[last].ofs)
                last = i;
        }
        
        if(last != 40)				// pak lump is not last in file - can't handle this situation
        {
            Cons.println("Pak lump is not last in file - cannot continue");
            return;
        }
        offset = lump[40].ofs;		// offset of pak file lump
        length = lump[40].len;		// length "   "   "    "
        
        if(length==0)
        {
            Cons.println("Map contains no pakfile!");
            return;
        }
        
        Cons.println("Pak file at offset "+offset+", length "+length+" bytes");
    }
    
    public void loadpak(RandomAccessFile bb) throws IOException	// load the pak file
    {
        
        // scan backwards from end to find End of Central Directory marker
        
        boolean cdirf = false;
        int off;
        for(off=offset+length-22; off>=0;off--)		// CD length is 22 bytes
        {
            bb.seek(off);
            
            int csig = Swab.I(bb.readInt());
            if(csig==0x06054b50)
            {
                cdirf=true; break;			// found it
            }
        }
        
        if(!cdirf)					// didn't find it
        {
            Cons.println("Couldn't find pak file EOCD");
            return;
        }
        
        bb.readShort();	//#of disk
        bb.readShort();	//#of disk with SOCD
        bb.readShort();	//#of disk with CD
        short cdes = Swab.S(bb.readShort());	//# of CD entries
        int cdsize = Swab.I(bb.readInt());	//CD
        int cdoffs = Swab.I(bb.readInt());
        bb.readShort();	// comment length
        
//        System.out.println("EOCD at "+(off-offset));
        Cons.println("Pak lump entries: "+cdes);
        
        if(cdes<=0) return;				// no files
        
        bb.seek(offset + cdoffs);			// to start of CD
//        System.out.println("CD at "+cdoffs);
        // read the zipfileheaders
        
        int [] zfcs = new int[cdes];
        int [] zfus = new int[cdes];
        int [] zfro = new int[cdes];
        int [] zfdo = new int[cdes];
        int [] zcrc = new int[cdes];
        String [] zffn = new String [cdes];
        
        for(int i=0;i<cdes;i++)				//ZipFileHeaders
        {
//            System.out.println("ZFH "+i+" at "+(bb.position()-offset));
            int hsig = Swab.I(bb.readInt());			//signature
            if(hsig != 0x02014b50)
            {
                Cons.println("ZipFH signature incorrect");
                return;
            }
            bb.readShort();	// vers made by
            bb.readShort();	// vers to extract
            bb.readShort();	// flags
            bb.readShort();	// comp meth
            bb.readShort();	// lastmod time
            bb.readShort();	// lastmod date
            zcrc[i] = Swab.I(bb.readInt());	// CRC32
            int csize = zfcs[i] = Swab.I(bb.readInt());	//comped size
            int usize = zfus[i] = Swab.I(bb.readInt());	//uncomped size
            short fnlen = Swab.S(bb.readShort());	//filename length
            short exlen = Swab.S(bb.readShort());	//extra field length
            short fclen = Swab.S(bb.readShort());	//file comment length
            bb.readShort();	// disk#start
            bb.readShort();	// int attrs
            bb.readInt();	// ext attrs
            int rolh = zfro[i] = Swab.I(bb.readInt());	// relative ofs of local header
            
            zffn[i] = readstr(bb,fnlen);
            readstr(bb,exlen+fclen);
            if(csize != usize)
            {
                Cons.println("Zip file "+i+" is compressed! "+csize+"!="+usize);
                return;
            }
            
      //			System.out.println("Size "+csize+"/"+usize+" fnl "+fnlen+" exl "+exlen+" fcl "+fclen+" RO "+rolh+" :\n"+zffn[i]);
            
        }
        //		System.out.println("\nLocal files:");
        
        for(int i=0;i<cdes;i++)			//ZipLocalFileHeaders
        {
            bb.seek(offset+zfro[i]);
  //          System.out.println("ZLFH "+i+" at "+(bb.position()-offset));
            
            int lsig = Swab.I(bb.readInt());
            if(lsig != 0x04034b50)
            {
                Cons.println("ZipLFH signature incorrect");
                return;
            }
            bb.readShort();	// version to extract
            bb.readShort();		// flags
            bb.readShort();	// compression method
            bb.readShort();		// last mod time
            bb.readShort();		// last mod date
            bb.readInt();		// CRC
            bb.readInt();		// Compressed size
            bb.readInt();		// Uncompressed size
            short fnlen = Swab.S(bb.readShort());		// filename length
            short exlen	= Swab.S(bb.readShort());		// extra field length
            zfdo[i]=zfro[i]+30+fnlen+exlen;
            
            //			System.out.println(readstr(bb,fnlen));
        }
        
        // now build the Zipfile arraylist
//JDK150        zf = new ArrayList<Zipf>();	// create an empty arraylist of zip files
        zf = new ArrayList();	// create an empty arraylist of zip files
        
        for(int i=0;i<cdes;i++)
        {
            Zipf z = new Zipf();	// current (i'th) zip file
            z.size=zfcs[i];		// file size
            z.relofs=zfro[i];		// local header relative offset
            z.datofs=zfdo[i];		// data offset
            z.CRC = zcrc[i];		// CRC32
            z.setfull(zffn[i]);	// fullname (also populate file & pathname)
            z.settype();		// sets the type depending on filename extension
            z.inpak = true;		// zip is in pak already
            z.data = null;		// data is in pak, stored already
            
            zf.add(z);		// add the zip to the zip file list
        }
    }
    public void writestr(RandomAccessFile b, String str) throws IOException // write string to buffer b
    {
        for(int i=0;i<str.length();i++)
        {
            b.writeByte(str.charAt(i));
        }
    }
    
    public String readstr(RandomAccessFile b, int len) throws IOException	// read len bytes from buffer b, return as string
    {
        StringBuffer linebuff = new StringBuffer();
        char c;
        for(int i=0;i<len;i++)
        {
            c=(char)b.readUnsignedByte();
            linebuff.append(c);
        }
        return linebuff.toString();
    }	//readstr
    
}

class Zipf		// a single zip file in the pak
{
    int size;		//compressed=uncompressed size
    int relofs;		// zip local header offset (into pak)
    int datofs;		// zip data offset (into pak)
    String fullname;	// path/file.ext
    String filename;	// file.ext
    String pathname;	// path only
    int type;		// 0=other, 1=material (vmt), 2=texture (vtf)
    boolean inpak;	// false if not yet added to this map's pak
    byte [] data;	// the zip file data, or NULL if already in pak
    long CRC;		// the zip's CRC
    
    public void setfull(String fn)	// set the full, file and path names from a fullname
    {
//	System.out.println("In ["+fn+"]");

	  if(fn.startsWith("/"))		// strip leading slash, if present
		fullname = fn.substring(1,fn.length());
        else
        	fullname = fn;

//	System.out.println("Mid ["+fullname+"]");


        int is = fullname.lastIndexOf("/");		// find the last slash (we will always uses forward slashes)
        if(is>=0 && is<fullname.length()-1)		// if position is valid or not last
        {
//		System.out.print("*");
            filename = fullname.substring(is+1);	// filename is everything after last slash
            pathname = fullname.substring(0,is);	// pathname is everything up to last slash
        }
        else	// probably pathless fullname
        {
//		System.out.print("!");
            filename = fullname;	
            pathname = "";		// empty path
        }
	  
 //	System.out.println("Out ["+pathname+"] ["+filename+"]");

    }
    
    public void setcfull(String fn)    // set the fullname, fixing-up backslash-to-slash
    {
        fn = fn.replace(File.separatorChar,'/');		// replace
        setfull(fn);							// set the fullname
    }
    
    public String getrelfull(String rootdir)        // returns fixed-up fullname, truncating path relative to likely base
    {
        String full = fullname;
        int index;
        if(!rootdir.equals(""))		// if gamedir pref not empty
        {
            if(full.startsWith(rootdir))	// fullname starts with the gamedir?
            {
                index = rootdir.length() + 1;	// then split off after the slash
                if(index<full.length())		// make sure index isn't too high
                    return full.substring(index);		// return all after gamedir+/
            }
        }			// no gamedir set or not a prefix, so guess root path from pathname
        
        index = full.indexOf("/materials");		// material file (vmt or vtf)
        if(index != -1)
            return full.substring(index+1);
        index = full.indexOf("/models");		// model file (mdl)
        if(index != -1)
            return full.substring(index+1);
        index = full.indexOf("/sound");		// sound file (wav)
        if(index != -1)
            return full.substring(index+1);
        return null;					// null if can't find a plausible relative path
    }
    
    public void setfile(String fn)	// set the filename, updating the fullname.
    {
//	System.out.println("SF ["+fn+"]");

        setfull(pathname+"/"+fn);
    }
    
    public void setpath(String pn)	// set the path, updating the fullname
    {
//	System.out.println("SP ["+pn+"]");

        setfull(pn+"/"+filename);
    }
    
    public void settype()	// set the file type, depending on the file extension
    {
        type = 0;		// default to other
        if(filename.toLowerCase().endsWith(".vmt")) type = 1;		// is a material file
        if(filename.toLowerCase().endsWith(".vtf")) type = 2;		// is a texture file
        // else is other
    }
    
    
}
