package pak;

import java.util.*;
import java.util.zip.*;
import java.util.prefs.*;
import java.io.*;
import java.nio.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.table.*;

public class Unpak
{
    
    Mappak m;	// the current map
    String currentdir; // current working directory
    File infile;			// current file
    RandomAccessFile raf;		// random access file handle
    
    Preferences prefs;	// prefs
    String gamedir;	// the game root directory
    int	fixup;	// 0=don't fixup , 1=ask to fixup, 2=auto fixup
    
    
    public int ubyte(byte b)
    {
        return (int) b & 0xFF;
    }
    
    public void exec(String filename)  throws Exception
    {
        try
        {
            
            currentdir = System.getProperty("user.dir");	// set current working directory
            
            prefs = Preferences.userRoot().node("Pakrat");	// our node
            gamedir = prefs.get("Gamedir", "");			// default to nothing
            fixup = prefs.getInt("Fixup", 1);			// default to ask
            //            System.out.println("Prefs: "+gamedir+" "+fixup);
            
            
//            UIManager.setLookAndFeel("com.sun.java.swing.plaf.windows.WindowsLookAndFeel");

		Cons.settitle("Pakrat - console");
		Cons.println("Pakrat v0.8 - by Rof (rof@mellish.org.uk)");

            if(filename==null)	// no file specified, use fileselector
            {
                if(!gamedir.equals(""))
                    currentdir = gamedir;
                
                JFileChooser chooser = new JFileChooser(currentdir);
                chooser.setDialogTitle("Open a map file");
                chooser.setFileFilter(new BspFileFilter());	// *.bsp files only
                int result = chooser.showOpenDialog(null);	// dialog with no parent container
                if (result == JFileChooser.CANCEL_OPTION)
                    return;						// user canceled
                
                infile = chooser.getSelectedFile( );	// file chosen
                filename = infile.getName();			// its name
            }
            else
            {
                infile = new File(filename);			// file from command line argument
            }
            
            if(!infile.exists() || !infile.canRead())
            {
                Cons.println("Can't open "+filename);
                return;
            }
            Cons.println("Reading "+filename);
            
            raf = new RandomAccessFile(infile,"r");
            
            //Read the BSP Header
            
            
            m = new Mappak();		//instance of the map
            m.loadmap(raf);		//load it
            
            final JFrame frame = new JFrame("Pakrat - " + filename);
            JPanel panel = new JPanel();
            panel.setLayout(new BorderLayout());
            
            JMenu filemenu = new JMenu("File");
            JMenuItem mload = new JMenuItem("Load BSP");
            JMenuItem msave = new JMenuItem("Save BSP");
            mload.setToolTipText("Load an new map file");
            msave.setToolTipText("Save the current map file");
            
            filemenu.add(mload);
            filemenu.add(msave);
            filemenu.addSeparator();
            JMenuItem mdump = new JMenuItem("Textures");
            filemenu.add(mdump);
            mdump.setToolTipText("List all textures/materials used in map world geometry");
            JMenuItem mpref = new JMenuItem("Preferences");
            mpref.setToolTipText("Set preferences");
            filemenu.add(mpref);
            filemenu.addSeparator();
            JMenuItem mquit = new JMenuItem("Quit");
            mquit.setToolTipText("Quit Pakrat");
            filemenu.add(mquit);
            
            JMenu helpmenu = new JMenu("Help");
            JMenuItem mhelp = new JMenuItem("About Pakrat");
            helpmenu.add(mhelp);
            
            JMenuBar menubar = new JMenuBar();
            menubar.add(filemenu);
            menubar.add(helpmenu);
            frame.setJMenuBar(menubar);
            
            
            
            
            final ZipDirModel model = new ZipDirModel(m.zf);
            model.setfileparams(raf, m.offset);
            final JTable table = new JTable(model);
            
            table.setAutoResizeMode(JTable.AUTO_RESIZE_NEXT_COLUMN);
            table.getColumn("Size").setMaxWidth(50);
            table.getColumn("Type").setMaxWidth(50);
            table.getColumn("In").setMaxWidth(20);
            table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
            
            panel.add(new JScrollPane(table),BorderLayout.CENTER);
            
            Container controls = Box.createHorizontalBox( );
            final JButton view = new JButton("View");
            view.setToolTipText("View the selected file contents");
            controls.add(view);
            final JButton editfile = new JButton("Edit");
            editfile.setToolTipText("Edit the selected file's path and filename");
            controls.add(editfile);
            controls.add(Box.createHorizontalStrut(30));
            final JButton addfile = new JButton("Add");
            addfile.setToolTipText("Add file(s) to the pak");
            controls.add(addfile);
            final JButton delfile = new JButton("Delete");
            delfile.setToolTipText("Delete the selected file from the pak");
            controls.add(delfile);
            final JButton savefile = new JButton("Save");
            savefile.setToolTipText("Save the selected file to disk");
            controls.add(savefile);
            
            panel.add(controls,BorderLayout.SOUTH);
            
            view.setEnabled(false);
            editfile.setEnabled(false);
            delfile.setEnabled(false);
            savefile.setEnabled(false);
            
            ListSelectionModel rowsel = table.getSelectionModel();	// listen for list selection changes
            rowsel.addListSelectionListener( new ListSelectionListener()
            {
                public void valueChanged(ListSelectionEvent lse)
                {
                    if (lse.getValueIsAdjusting()) return;	// not a selection event
                    ListSelectionModel lsm = (ListSelectionModel)lse.getSource();
                    
                    if(lsm.isSelectionEmpty() )			// no selected row
                    {
                        view.setEnabled(false);		// disable button
                        editfile.setEnabled(false);
                        delfile.setEnabled(false);
                        savefile.setEnabled(false);
                    }
                    else
                    {						//enable buttons
                        view.setEnabled(true);
                        editfile.setEnabled(true);
                        delfile.setEnabled(true);
                        savefile.setEnabled(true);
                    }
                }
            });
            
            // MENU AND BUTTON ACTION HANDLERS
            
            msave.addActionListener(new ActionListener()		// handler for Save BSP menu item
            {	//save the bsp, updating the pak
                public void actionPerformed(ActionEvent e)
                {
                    try
                    {
                        File sfile = new File(infile.getAbsolutePath());	// copy of current file spec
                        ZipDirModel tmodel = (ZipDirModel)table.getModel();   // the table model
                        
                        JFileChooser schooser = new JFileChooser(sfile);
                        schooser.setDialogTitle("Save map file - "+sfile.getName());
                        schooser.setFileFilter(new BspFileFilter());	// *.bsp files only
                        schooser.setSelectedFile(sfile);			// select the save file, if we can
                        int result = schooser.showSaveDialog(frame);
                        if (result == JFileChooser.CANCEL_OPTION)
                            return;						// user canceled
                        
                        sfile = schooser.getSelectedFile( );	// file chosen
                        String sfilename = sfile.getName();			// its name
                        
                        // if a file already exist, we want to ask whether to overwrite it
                        // if that file is the current file, we need to rename it
                        // actually make a copy of the current file since renaming seems dodgy
                        // now re-written to avoid nio totally
                        //
                        //
                        // fail if user chooses not to overwrite
                        // or if we can't copy the file
                        // or if once copied, we can't find or reopen a buffer to it
                        
                        
                        if(sfile.exists())		// does file exist?
                        {					// check if user wants to overwrite
                            result = JOptionPane.showConfirmDialog(frame,
                            "Map file \""+sfile+"\" exists. \nAre you sure you want to overwrite?", "Save BSP file", JOptionPane.YES_NO_OPTION);
                            if(result == JOptionPane.NO_OPTION)	// user cancelled
                                return;				// so don't write
                            
                            if(sfile.getCanonicalPath().equals(infile.getCanonicalPath()))
                            {	// we are overwriting currently open map file, so rename it
                                // by copying it to a new file
                                
                                long ilength = infile.length();	// size of the current file
                                
                                File renfile = new File(infile.getAbsolutePath()+".bak"); // this is the new name
                                
                                Cons.print("Copying current map file to "+renfile.getAbsolutePath()+"...");
                                
                                RandomAccessFile copyraf = new RandomAccessFile(renfile, "rw");     // the copy of the current file
                                copyraf.setLength(0);                // if it already exists, zero it's length
                                                 
                                raf.seek(0);                        // seek to beginning of current file
                                for(int i=0;i<ilength;i++)          // for each byte
                                {
                                    copyraf.write(raf.readUnsignedByte());      // read-write
                                }
                                copyraf.close();            // close the copy
                                
                                Cons.println("Done");
                                
                                infile = renfile;       // the current file is now the renamed file
                                if(!infile.exists())	// check for correct copy
                                {
                                    Cons.println("Cannot find renamed file - map save aborted");
                                    return;		// didn't work
                                }
                                // it worked, so update all file params to point to the renamed file
                                
                                raf.close();        // close the current file
                                raf = new RandomAccessFile(infile,"r");		// update the current file to the copy
                                
                                tmodel.setfileparams(raf, m.offset);	// update the model's file params
                            }
                        }
                        
                                /*	if(!sfile.canWrite())		//this doesn't work?
                                {
                                        Cons.println("Can't write file: "+sfile);
                                        return;
                                }	*/
                        
                        
                        
                        
                        Cons.print("Writing "+sfilename+"...");
                        
                        
                        raf.seek(0);		// seek to beginning of current file
                        
                        
                        
                        //open a byte buffer for writing
                        RandomAccessFile outraf = new RandomAccessFile(sfile,"rw");
                        outraf.setLength(0);    // wipe the file, if it exists already
                        Cons.print("BSP data...");
                        
                        m.savemap(raf,outraf);	// write data up to pak lump
                        
                        Cons.print("Pak data...");
                        
                        m.savepak(raf,outraf);	// write the pak lump
                        
                        
                        outraf.close();		// all done
                        
                        Cons.println("Done");
                        
                        // now update current file ops to point to the just-saved file
                        
                        raf.close();		// close the old file
                        
                        infile = sfile;		// current file is save file
                        raf = new RandomAccessFile(infile,"r");		// update the current file
                        
                        // dont need to re-load it, however
                        
                        // saving the BSP updates stored entries to in-pak entries, so update the table
                        tmodel.setfileparams(raf, m.offset);	// update the model's file params
                        tmodel.fireTableDataChanged();	// force update of table view
                        
                        frame.setTitle("Pakrat - " + sfile.getName());	// set window title
                        
                        
                        
                    }
                    catch(Exception ex)
                    {
                        System.out.println(ex);
                    }
                    
                }
            });
            
            
            savefile.addActionListener(new ActionListener()		// handler for save file button
            {									//save a file from the pak list to disk
                public void actionPerformed(ActionEvent e)
                {
                    ZipDirModel tmodel = (ZipDirModel)table.getModel();
                    int off = tmodel.getoffset();
                    int row = table.getSelectedRow();		// the selected file row index
                    if(row==-1)		// no selection
                        return;	// so cannot save
                    Zipf z = tmodel.getzipfile(row);		// this is the zip file to be saved
                    
                    ByteBuffer zb;
                    
                    try
                    {
                    
                    
                    if(z.inpak)         // file is already in pak
                    {
                        raf.seek(off+z.datofs);  //seek to start of the zip data
                        byte [] buffer = new byte[z.size];  // a buffer to hold the zip file
                        raf.read(buffer);                   // read the data into the buffer
                        zb = ByteBuffer.wrap(buffer);   // wrap the loaded data in a bytebuffer
                    }
                    else                // file not in pak
                    {
                        zb = ByteBuffer.wrap(z.data);   // wrap the stored data in a bytebuffer
                    }
                       
                    zb.order( ByteOrder.LITTLE_ENDIAN ) ;	// set endianness
                    
                        
                        File sfile = new File(z.fullname);	// the file to save
                        // currently relative path, so will fail to match
                        // Do plausible pathfinding later?
                        
                        JFileChooser schooser;
                        if(gamedir.equals(""))			// for now, set dir to gamedir, if pref is set
                            schooser = new JFileChooser(sfile);
                        else
                            schooser = new JFileChooser(gamedir);
                        
                        schooser.setDialogTitle("Save selected file - "+z.fullname);
                        schooser.setSelectedFile(sfile);			// select the save file, if we can
                        int result = schooser.showSaveDialog(frame);
                        if (result == JFileChooser.CANCEL_OPTION)
                            return;						// user cancelled the save
                        
                        sfile = schooser.getSelectedFile( );	// file chosen
                        String sfilename = sfile.getName();		// its name
                        
                        if(sfile.exists())		// does file exist?
                        {					// check if allowed to overwrite
                            result = JOptionPane.showConfirmDialog(frame,
                            "File \""+sfile+"\" exists, overwrite?", "Save selected file", JOptionPane.YES_NO_OPTION);
                            if(result == JOptionPane.NO_OPTION)	// user cancelled
                                return;					// so don't write
                        }
                        
                        Cons.print("Writing "+sfilename+" ...");
                        
                        FileOutputStream fos = new FileOutputStream(sfile);	// stream to write
                        
                        for(int i=0;i<z.size;i++)			// for length of file
                        {
                            fos.write((int)zb.get());			// write bytes from buffer to file
                        }
                        fos.close();					// all done
                        Cons.println("Done");
                        
                        
                    }
                    catch(Exception ex)
                    {
                        System.out.println(ex);
                    }
                    
                }
            });
            
            
            
            addfile.addActionListener(new ActionListener()		// handler for add file button
            {    							        	//add a file to the pak list
                public void actionPerformed(ActionEvent e)
                {
                    
                    try
                    {
                        ZipDirModel tmodel = (ZipDirModel)table.getModel();
                        
                        JFileChooser fchooser = new JFileChooser(currentdir);	// locate the file
                        fchooser.setDialogTitle("Select file(s) to add to pak");
                        fchooser.setApproveButtonToolTipText("Open the selected files(s)");
                        fchooser.setFileFilter(new PakFileFilter());	// *.bsp files only
                        fchooser.setMultiSelectionEnabled(true);        // allow multi-file selection
                        int result = fchooser.showOpenDialog(frame);
                        if (result == JFileChooser.CANCEL_OPTION)
                            return;						// user canceled
                        
                        File [] tfile = fchooser.getSelectedFiles(); // get the selected files
                        
                        for(int i=0;i<tfile.length;i++)        // loop over all files chosen
                        {
                            String tfilename = tfile[i].getName();	// i'th file's name
                            
                            if(!tfile[i].exists() || !tfile[i].canRead())	// file ok?
                            {
                                Cons.println("Can't open "+tfilename);
                                return;
                            }
                            Cons.println("Reading "+tfilename);
                            currentdir = tfile[i].getPath();			// update the current directory
                            Zipf z = new Zipf();					// this will be the new zip file
                            z.setcfull(tfile[i].getAbsolutePath());  // set fullname with slash fixup
                            
                            if(fixup != 0)	// if preference isn't never-fixup
                            {
                                String relfull = z.getrelfull(gamedir);	// see if path can be truncated to known relative path
                                if(relfull!=null)				// is so
                                {
                                    if(fixup == 1)	// preference is to ask
                                    {
                                        result = JOptionPane.showConfirmDialog(frame,
                                        z.fullname+"\nFix-up path to: \""+relfull+"\" ?", "File "+(i+1)+" of "+tfile.length, JOptionPane.YES_NO_OPTION);
                                        if(result == JOptionPane.YES_OPTION)	// user agreed
                                            z.setfull(relfull);			// set to the relative path
                                    }
                                    else			// preference is auto fixup
                                        z.setfull(relfull);	// fixit
                                }
                            }
                            
                            z.size = (int) tfile[i].length();	// zip size is file length
                            z.datofs = z.relofs = 0;		// offsets not set
                            z.settype();				// autoset the filetype from the extension
                            z.inpak = false;				// is not in the pak
                            
                            z.data = new byte[z.size];		// allocate and read the data
                            FileInputStream fis = new FileInputStream(tfile[i]);
                            fis.read(z.data,0,z.size);		// read into the buffer
                            fis.close();
                            
                            CRC32 crc = new CRC32();		// compute the CRC
                            crc.update(z.data);			// from the data buffer
                            z.CRC = crc.getValue();			// store it for the zip file
                            
                            
                            if(m.zf == null)				// deal with empty pak
                            {
//JDK150                                m.zf = new ArrayList<Zipf>();	// create a new list
                                m.zf = new ArrayList();	// create a new list
                                tmodel.setziplist(m.zf);	// and update model to use it
                            }
                            tmodel.addfile(z);			// add the new zip file to the pak list
                        }
                        
                    }
                    catch(Exception ex)
                    {
                        System.out.println(ex);
                    }
                    
                }
            });
            
            delfile.addActionListener(new ActionListener()		// handler for delete file button
            { 							           //delete the selected file from the pak list
                public void actionPerformed(ActionEvent e)
                {
                    int row = table.getSelectedRow();			// determine selected file
                    if(row==-1)
                        return;         					//no selection
                    ZipDirModel tmodel = (ZipDirModel)table.getModel();
                    Zipf z = tmodel.getzipfile(row);			// this is the selected zip file
                    // confirm deletion
                    int result = JOptionPane.showConfirmDialog(frame, "Remove file "+z.filename+" from the pak?",
                    "Delete file", JOptionPane.YES_NO_OPTION);
                    if(result == JOptionPane.NO_OPTION)
                        return;             // user cancelled
                    
                    tmodel.deletefile(row);				// remove from pak list
                }
            });
            
            editfile.addActionListener(new ActionListener()		// handler for file edit button
            {									// edit the file properties
                public void actionPerformed(ActionEvent e)
                {
                    ZipDirModel tmodel = (ZipDirModel)table.getModel();
                    int row = table.getSelectedRow();
                    if(row==-1)		//no selection
                        return;
                    Zipf z = tmodel.getzipfile(row);
                    
                    
                    JTextField filetext = new JTextField(z.filename);		// filename edit box
                    JTextField pathtext = new JTextField(z.pathname);		// pathname edit box
                    Container cbox = Box.createHorizontalBox( );
                    cbox.add( new JLabel("Size: "+z.size+"  CRC32: "+Integer.toHexString( (int)z.CRC )) );
                    
                    Container fbox = Box.createHorizontalBox( );
                    fbox.add(new JLabel("Filename : ")); fbox.add(filetext);
                    Container pbox = Box.createHorizontalBox( );
                    pbox.add(new JLabel("Path : ")); pbox.add(pathtext);
                    
                    int result = JOptionPane.showOptionDialog(frame,		// edit dialogue
                    new Object[] { z.fullname, cbox, fbox, pbox },
                    "Edit file parameters", JOptionPane.OK_CANCEL_OPTION,
                    JOptionPane.PLAIN_MESSAGE, null, null, null);
                    
                    if(result == JOptionPane.CANCEL_OPTION)	return;	// user cancelled
                    
                    z.setfull( pathtext.getText() + "/" + filetext.getText() );	// update fullname
                    z.settype();									// and type
                    tmodel.fireTableDataChanged();						// update table
                    
                }
            });
            
            mpref.addActionListener(new ActionListener()			// handler for preferences menu item
            {										// preferences dialogue
                public void actionPerformed(ActionEvent e)
                {
                    final JTextField gdirtext = new JTextField(gamedir);	// this is the root game dir
                    gdirtext.setPreferredSize(new Dimension(200,-1));
                    gdirtext.setToolTipText("The directory to strip off path of files added to the pak");
                    JButton gdfind = new JButton("...");
                    gdfind.setToolTipText("Browse to set the game root directory");
                    
                    Container gbox = Box.createHorizontalBox( );
                    gbox.add(new JLabel("Game root directory :   ")); gbox.add(gdirtext);
                    gbox.add(gdfind);
                    
                    JComboBox fcombo = new JComboBox(new String [] {"Never","Ask","Always" } );
                    fcombo.setSelectedIndex(fixup);
                    fcombo.setToolTipText("Sets whether to strip off the root directory on adding a file");
                    Container fbox = Box.createHorizontalBox( );
                    fbox.add(new JLabel("Path fixup on add file :   ")); fbox.add(fcombo);
                    
                    
                    gdfind.addActionListener(new ActionListener()
                    {		// popup dir chooser
                        public void actionPerformed(ActionEvent ae)
                        {
                            JFileChooser dc = new JFileChooser(gdirtext.getText());
                            dc.setDialogTitle("Set game root directory");
                            dc.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
                            int dr = dc.showOpenDialog(frame);
                            if (dr != JFileChooser.CANCEL_OPTION)
                                gdirtext.setText(dc.getSelectedFile().getPath());
                        }
                    });
                    
                    // show the dialogue
                    int result = JOptionPane.showOptionDialog(frame,
                    new Object[] {  gbox, fbox },
                    "Preferences", JOptionPane.OK_CANCEL_OPTION,
                    JOptionPane.PLAIN_MESSAGE, null, null, null);
                    
                    if(result == JOptionPane.CANCEL_OPTION)	return;	// user cancelled
                    
                    gamedir = gdirtext.getText();		// the new gamedir
                    gamedir = gamedir.replace(File.separatorChar,'/'); // backslash/slash conversion
                    if(gamedir.endsWith("/"))		// kill last slash if present (usually only for X:/)
                        gamedir = gamedir.substring(0,gamedir.length()-1);
                    
                    fixup = fcombo.getSelectedIndex();	// the new fixup-mode
                    prefs.put("Gamedir", gamedir);	// write them to prefs
                    prefs.putInt("Fixup", fixup);
                }
            });
            
            mload.addActionListener(new ActionListener()			// handler for load BSP menu item
            {										// load a new BSP file
                public void actionPerformed(ActionEvent e)
                {
                    try
                    {
                        JFileChooser rchooser = new JFileChooser(currentdir);
                        rchooser.setDialogTitle("Open a map file");
                        rchooser.setFileFilter(new BspFileFilter());	// *.bsp files only
                        int result = rchooser.showOpenDialog(frame);
                        if (result == JFileChooser.CANCEL_OPTION)
                            return;						// user canceled
                        
                        
                        
                        infile = rchooser.getSelectedFile( );	// file chosen
                        String filename = infile.getName();			// its name
                        currentdir = infile.getPath();
                        
                        if(!infile.exists() || !infile.canRead())	// check for existance and readability
                        {
                            Cons.println("Can't open "+filename);
                            return;
                        }
                        Cons.println("Reading "+filename);
                        
                        raf = new RandomAccessFile(infile,"r");		// update the current file and buffers
                        
                        m = new Mappak();		//new instance of the map
                        m.loadmap(raf);		//load it
                        frame.setTitle("Pakrat - " + filename);	// set window title
                        
                        ZipDirModel nmodel = new ZipDirModel(m.zf);		// update the table model
                        nmodel.setfileparams(raf, m.offset);				// and the model's file params
                        table.setModel(nmodel);						// reset the table
                        table.setAutoResizeMode(JTable.AUTO_RESIZE_NEXT_COLUMN);
                        table.getColumn("Size").setMaxWidth(50);
                        table.getColumn("Type").setMaxWidth(50);	// reset the column options
                        table.getColumn("In").setMaxWidth(20);	// since they seem to default on model change
                        
                    } catch (Exception ex)
                    { System.out.println(ex); }
                    
                }
            });
            
            mquit.addActionListener(new ActionListener()		// handler for Quit menu item
            {		// Quit application
                public void actionPerformed(ActionEvent e)
                {									// confirm quit action
                    int result = JOptionPane.showConfirmDialog(frame, "Quit Pakrat?",
                    "Pakrat", JOptionPane.YES_NO_OPTION);
                    if(result == JOptionPane.YES_OPTION)
                        System.exit(0);					// exit the app
                    
                }
            });
            
            
            view.addActionListener(new ActionListener()		// handler for view file button
            {									// show the selected file
                public void actionPerformed(ActionEvent e)
                {
                    ZipDirModel tmodel = (ZipDirModel)table.getModel();
                    int off = tmodel.getoffset();
                    int row = table.getSelectedRow();
                    if(row==-1)		//no selection
                        return;
                    Zipf z = tmodel.getzipfile(row);			// this zip file selected
                    
                    ByteBuffer zb;
                    
                    try {
                    
                    if(z.inpak)         // file is already in pak
                    {
                        raf.seek(off+z.datofs);  //seek to start of the zip data
                        byte [] buffer = new byte[z.size];  // a buffer to hold the zip file
                        raf.read(buffer);                   // read the data into the buffer
                        zb = ByteBuffer.wrap(buffer);   // wrap the loaded data in a bytebuffer
                    }
                    else                // file not in pak
                    {
                        zb = ByteBuffer.wrap(z.data);   // wrap the stored data in a bytebuffer
                    }
                       
                    zb.order( ByteOrder.LITTLE_ENDIAN ) ;	// set endianness
                    
                    String text;
                    
                    switch(tmodel.gettype(row))			// determine what format to show, given the file type
                    {
                        case 0:	//other or unknown - hex dump
                        default:
                            text = hexlist( readstr(zb, z.size) );
                            break;
                        case 1:	//material (vmt) - text dump
                            text = readstr(zb, z.size);
                            break;
                        case 2:	//texture (vtf) - texture info
                            text = vtfinfo(zb, z.size);
                            break;
                    }
                    
                    TextBox("Pakrat - "+z.fullname, text);	// display the textbox
                    } catch(Exception ex)
                    { System.out.println(ex);
                    }
                }
            });
            
            mdump.addActionListener(new ActionListener()			// handler Texture menu item
            {										// list all textures used in world
                public void actionPerformed(ActionEvent e)
                {
                    ZipDirModel tmodel = (ZipDirModel)table.getModel();
                    
                    StringBuffer t = new StringBuffer("Material list\n");
                    
                    int sofs = m.lump[43].ofs;	// TexDataStringData Lump
                    int slen = m.lump[43].len;	// has actual texture name character data
                    
                    try {
                    
                    raf.seek(sofs);			// start of TDSD
                    String tdsd = readstr(raf, slen);
                    
                    int ofs = m.lump[44].ofs;		// TexDataStringTable Lump
                    int len = m.lump[44].len;		// has relative pointers to each texture name string
                    int numtdst = len/Lump.size(44);
                    
                    int [] tdst = new int[numtdst];	// array of string offsets
                    String [] texname = new String[numtdst];	// array of texture strings
                    
                    raf.seek(ofs);			// start of TDST lump
                    for(int i=0;i<numtdst;i++)		// build string array from null-terminated strings
                    {
                        int ix = tdst[i] = Swab.I(raf.readInt());	// start of texstring
                        
                        for(int j=ix;j<sofs;j++)		// scan forward
                        {
                            if(tdsd.charAt(j)==0)		// look for null terminator
                            {
                                texname[i]=tdsd.substring(ix,j);	//end of texstring
                                break;				// so look for next string
                            }
                        }
                        t.append(texname[i]+"\n");		// append found texstring to list
                    }
                    
                    
/*			// temporary list of pak entries
                          t.append("\n-------------\n\n");
 
                        int numzips = m.zf.size();
                        for(int i=0;i<numzips;i++)
                        {
                                Zipf z = m.zf.get(i);
                                t.append(i+": "+z.size+" "+z.relofs+" "+z.datofs+" "+z.type+" "+Integer.toHexString((int)z.CRC)+" "+z.inpak+"\n");
                                t.append(z.fullname+"\n");
                        } */
                    
                    TextBox("Pakrat - materials - "+ infile.getName(), t.toString() );	// display resulting list
                    } catch (Exception ex)
                    {
                        System.out.println(ex);
                    }
                }
            });
            
            mhelp.addActionListener(new ActionListener()		// handler for help menu item
            {									// just a box of text
                public void actionPerformed(ActionEvent e)
                {
                    String help =" Pakrat v0.8 -- by Rof (rof@mellish.org.uk)\n\n"
                    +" A program for managing Half-Life 2 BSP PAK archives\n\n"
                    +" File menu:\n"
                    +" 	Load BSP 	- load a new BSP file\n"
                    +" 	Save BSP 	- save the current BSP file, writing any changes to pak\n"
                    +" 	Textures 	- list material/texture files used in map geometry\n"
                    +"	Preferences	- set the game base directory and path-fixup options\n\n"
                    +" Help menu:\n	About Pakrat - this information\n\n"
                    +" Button controls:\n"
                    +"	View		- view the selected pak entry\n"
                    +"	Edit		- edit the selected entry's file and path name\n"
                    +"	Add			- add a file or files to the pak\n"
                    +"	Delete		- delete the selected entry from the pak\n"
                    +"	Save		- save the selected entry to disk\n\n"
                    +" About Pakrat\n"
                    +" HL2 map (.bsp) files contain a general file storage area, known as the\n"
                    +" pak. Usually this area contains special material (.vmt) and texture\n"
                    +" (.vtf) files which store the environment reflection maps from\n"
                    +" env_cubemap entities generated when the console command buildcubemaps\n"
                    +" is run. These files will be visible in the pak list of opened maps.\n\n"
                    +" Pakrat allows you to add files to the pak, such as texture, material,\n"
                    +" sound and model files. If these files are used in the map, they will\n"
                    +" be preferentially loaded from the map's pak, allowing you to make\n"
                    +" maps with custom textures, etc., embedded into the map .bsp file.\n"
                    +" These maps therefore do not need to be distributed with extra files\n"
                    +" to include custom components.\n\n"
                    +" Path fixup\n"
                    +" The Source engine looks for files in the pak with a certain relative\n"
                    +" paths. For example, material and texture files should have a path\n"
                    +" starting with the \"materials\" folder. If set to do so, Pakrat can\n"
                    +" attempt to change the path of any file added to the pak such that it\n"
                    +" is correct. The best way to do this is set the Game Root directory\n"
                    +" under the Preferences menu item. This should be, for example:\n"
                    +" \"C:\\Games\\Steam\\SteamApps\\<your steam name>\\half life 2\\hl2\"\n"
                    +" for a typical HL2 installation. If mapping for CS:S or HL2DM, change\n"
                    +" the Game Root appropriately. If the Game Root is not set, Pakrat\n"
                    +" can attempt to guess the correct path from the file name and location.\n"
                    +" You may also edit the pak entry's file and path directly, using\n"
                    +" the Edit button.\n\n"
                    +" The View button shows the contents of the selected file. For material\n"
                    +" (.vmt) files, the file is displayed as text. For textures (.vtf), a\n"
                    +" summary of the texture properties is printed. Unrecognised file types\n"
                    +" are displayed as a hex dump.\n"
                    
                    
                    ;
                    
                    
                    
                    TextBox("About Pakrat", help );			// display the text
                }
            });
            
            
            frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
            frame.setSize(640,480);
            frame.getContentPane().add(panel);
            frame.setVisible(true);

		            
        } catch(Exception e)
        { System.out.println(e); return; }
    }
    // exec
    
    public void TextBox(String title, String text)		// add a text frame with given title and content
    {
        JTextArea textarea = new JTextArea(text);
        textarea.setFont(new Font("Monospaced",Font.PLAIN,14));
        textarea.setTabSize(4);
        textarea.setEditable(false);
        
        JFrame tframe = new JFrame(title);
        tframe.setSize(600,250);
        tframe.getContentPane().add(new JScrollPane(textarea));
        tframe.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);	// will delete itself if closed
        tframe.setVisible(true);
    }
    
    
    
    
    public static void main(String args[]) throws Exception		// application entry point
    {
        String fn;
        if(args.length != 1)	//if have argument, use it as filename
            fn=null;		//else pass null for fileselector
        else
            fn=args[0];
        
        Unpak inst = new Unpak();		// instance of our application
        inst.exec(fn);				// execute it
        return;
    }
    
    public String readstr(RandomAccessFile r, int len) throws IOException	// read len bytes from RAF r, return as string
    {
        StringBuffer linebuff = new StringBuffer();
        char c;
        for(int i=0;i<len;i++)
        {
            c=(char)r.readUnsignedByte();
            linebuff.append(c);
        }
        return linebuff.toString();
    }	//readstr
    
    public String readstr(ByteBuffer b, int len)	// read len bytes from ByteBuffer b, return as string
    {
        StringBuffer linebuff = new StringBuffer();
        char c;
        for(int i=0;i<len;i++)
        {
            c=(char)b.get();
            linebuff.append(c);
        }
        return linebuff.toString();
    }	//readstr
    
    public String vtfinfo(ByteBuffer b, int size)			// parse VTF header and return info as string
    {
        StringBuffer t = new StringBuffer();
        
        char[] type = new char[4];
        for(int i=0;i<4; i++) type[i] = (char)b.get();
        
        String tstr = new String(type);
        
        if(!tstr.equals("VTF\00"))			// check format is understood
        {
            t.append("Invalid VTF file\n");
            return t.toString();
        }
        t.append("VTF file "+size+" bytes\n");
        
        int[]	vers = new int[2];
        vers[0] = b.getInt();
        vers[1] = b.getInt();
        t.append("Version "+vers[0]+"."+vers[1]+"\n");		// version info
        
        b.getInt();	//headersize
        short	width = b.getShort();
        short height = b.getShort();
        int flags = b.getInt();
        
        t.append("WxH: "+width+" x "+height+"\n");
        // texture flags (bit field)
        String []flagstr =
        {"POINTSAMPLE", "TRILINEAR","CLAMP-S", "CLAMP-T", "ANISOTROPIC","HINT-DXT5","NOCOMPRESS",
         "NORMAL", "NOMIP", "NOLOD", "MINMIP", "PROC", "1BALPHA", "8BALPHA", "ENVMAP",
         "RENDERTARGET", "DEPTH-RT", "NODEBUGOVERRIDE", "SINGLECOPY", "1OVERMIPLEVELINALPHA",
         "PREMULTCOL1OML", "NORMALTODUDV" };
         
         t.append("Flags:");
         for(int i=0;i<22;i++)					// determine which flags are set
         {
             if((flags & 0x1) ==1)
                 t.append(" "+flagstr[i]);
             flags >>= 1;
         }
         t.append("\n");
         
         int numframes = b.getShort();
         int startframe = b.getShort();
         t.append("Start "+startframe+" of "+numframes+" frames\n");
         
         b.getInt(); //pad					padding due to float alignment to 16-byte boundaries
         float refx = b.getFloat();
         float refy = b.getFloat();
         float refz = b.getFloat();
         b.getInt(); //pad
         float bumpscale = b.getFloat();
         t.append("Reflectivity: "+refx+", "+refy+", "+refz+"   Bumpscale: "+bumpscale+"\n");
         // image format names
         String imgfmt[] =
         {"RBGA8888", "ABGR8888", "RGB888", "BGR888", "RGB565", "I8", "IA88", "P8", "A8",
          "RBG888-BS", "BGR888-BS", "ARGB8888"," BGRA8888", "DXT1", "DXT3", "DXT5",
          "BGRX8888", "BGR565", "BGRX5551", "BGRA4444", "DXT1_1BA", "BGRA5551",
          "UV88", "UVWQ8888", "RGBA16161616F", "RGBA16161616" };
          
          
          int imageformat = b.getInt();
          int nummips = b.get();
          int lrimageformat = b.getInt();
          int lrwidth = b.get();
          int lrheight = b.get();
          if(imageformat<imgfmt.length)
              t.append("Image format: "+imgfmt[imageformat]);
          else
              t.append("Unknown image format: "+imageformat);
          
          t.append(" with "+nummips+" mip levels\n");
          
          if(lrimageformat>0)
          {
              if(lrimageformat<imgfmt.length)
                  t.append("Low-res format: "+imgfmt[lrimageformat]);
              else
                  t.append("Unknown lrif: "+lrimageformat);
              
              t.append(" WxH: "+lrwidth+" x "+lrheight+"\n");
          }
          return t.toString();
    }	// vtfinfo
    
    
    public String hexlist(String input)			// return string as hexdump (for unknown file types)
    {
        StringBuffer text = new StringBuffer();
        
        for(int i=0;i<input.length();i++)
        {
            if(i%32 == 0) text.append("\n");
            int c = (int)input.charAt(i) & 0xFF;
            if(c<16) text.append("0");
            text.append(Integer.toHexString(c)+" ");
        }
        
        return text.toString();
    } //hexlist
    

    //Horrible voodoo coding to force garbage collection of a MappedByteBuffer
    
    // needed until some form of unmap() exists for MBBs
    // else all other file ops fail on a previously-mapped file
    /*
    public static void clean(final Object buffer) throws Exception
    {
        AccessController.doPrivileged(new PrivilegedAction()
        {
            public Object run()
            {
                try
                {
                    Method getCleanerMethod = buffer.getClass().getMethod("cleaner",
                    new Class[0]);
                    getCleanerMethod.setAccessible(true);
                    sun.misc.Cleaner cleaner =
                    (sun.misc.Cleaner)getCleanerMethod.invoke(buffer,new Object
                    [0]);
                    cleaner.clean();
                } catch(Exception e)
                {
                    e.printStackTrace();
                }
                return null;
            }
        });
    }
    */

} //unpak


